From Coq Require Import ssreflect ssrbool ssrfun.
From fcsl Require Import axioms pred prelude pcm.
From fcsl Require Import notation state morphism invertible.
From fcsl Require Import options.

Section MorphismsAndCAP.
Variable U V : pcm.
Variable T : Type.
Variable D : sep_rel U.
Variable f : morphism V D.

Definition R : Pred (state U T) := 
  fun s => D (idfun.l s) (idfun.r s).

Lemma duplicable (s : state U T) :
        valid (idfun.x s) -> inv_rel D -> 
        s \In R <-> s \In star R R.
Proof.
move=>V1; rewrite -!toPredE /R /= => HD; split=>E; last first.
- case: E V1=>s1 [s2][]; rewrite /splits !lT -!toPredE /=.
  by case=>->->-> _ _ H1 H2 V1; case/andP: (HD _ _ _ V1 H1 H2).  
exists (State (idfun.l s) (joint s) (idfun.r s)).
exists (State Unit (joint s) (idfun.x s)). 
by rewrite /splits unitL unitR !lT -!toPredE /= ?lE ?allvE // (sep0U _ E).
Qed.

Definition F b : Pred (state U T) := 
  fun s => f.l s = b /\ D (idfun.l s) (idfun.r s).

Lemma homomorphic (s : state U T) b1 b2 :
        valid (idfun.x s) -> inv_morph f -> 
        s \In F (b1 \+ b2) <-> s \In star (F b1) (F b2).
Proof.
move=>W [HD Hf]; split=>[[/= E X]|[s1][s2][S]]; last first.
- case: S (W); rewrite /F -!toPredE /= -!lT =>->->-> _ _; rewrite !lT.
  move=>W1 [F1 X1][F2 X2]; rewrite pfjoin ?(validLE3 W1) // ?F1 ?F2;
  by case/andP: (HD _ _ _ W1 X1 X2). 
case/(Hf _ _ _ (validL W) (sepx0 W X)): E=>a1 [a2][/= E _ Y F1 F2].
exists (State a1 (joint s) (a2 \+ idfun.r s)).
exists (State a2 (joint s) (a1 \+ idfun.r s)).
by rewrite E in X W; rewrite /splits !lT /F -!toPredE /= !lE !(sepAxx W Y X).
Qed.

End MorphismsAndCAP.
